@@ -3,11 +3,11 @@ require File.expand_path('../helper', __FILE__)
 begin
 require 'liquid'
 
-class LiquidTest < Test::Unit::TestCase
+class LiquidTest < Minitest::Test
   def liquid_app(&block)
     mock_app do
       set :views, File.dirname(__FILE__) + '/views'
-      get '/', &block
+      get('/', &block)
     end
     get '/'
   end
@@ -42,18 +42,36 @@ class LiquidTest < Test::Unit::TestCase
 
   it "raises error if template not found" do
     mock_app { get('/') { liquid :no_such_template } }
-    assert_raise(Errno::ENOENT) { get('/') }
+    assert_raises(Errno::ENOENT) { get('/') }
   end
-  
+
   it "allows passing locals" do
-    liquid_app do
+    liquid_app {
       liquid '{{ value }}', :locals => { :value => 'foo' }
-    end
+    }
     assert ok?
     assert_equal 'foo', body
   end
+
+  it "can render truly nested layouts by accepting a layout and a block with the contents" do
+    mock_app do
+      template(:main_outer_layout) { "<h1>Title</h1>\n{{ yield }}" }
+      template(:an_inner_layout) { "<h2>Subtitle</h2>\n{{ yield }}" }
+      template(:a_page) { "<p>Contents.</p>\n" }
+      get('/') do
+        liquid :main_outer_layout, :layout => false do
+          liquid :an_inner_layout do
+            liquid :a_page
+          end
+        end
+      end
+    end
+    get '/'
+    assert ok?
+    assert_body "<h1>Title</h1>\n<h2>Subtitle</h2>\n<p>Contents.</p>\n"
+  end
 end
 
 rescue LoadError
-  warn "#{$!.to_s}: skipping liquid tests"
+  warn "#{$!}: skipping liquid tests"
 end
